# -*- coding: utf-8 -*-

"""Utilities for web applications."""

import importlib
from typing import Any, Mapping, NoReturn, Optional, TYPE_CHECKING, Union

import click

from .options import debug_option, host_option, port_option, verbose_option, with_gunicorn_option, workers_option

if TYPE_CHECKING:
    import flask  # noqa
    import gunicorn.app.base  # noqa

__all__ = [
    'make_web_command',
    'run_app',
    'make_gunicorn_app',
]


def make_web_command(
    app: Union[str, 'flask.Flask'],
    *,
    group: Optional[click.Group] = None,
    command_kwargs: Optional[Mapping[str, Any]] = None,
) -> click.Command:
    """Make a command for running a web application"""
    if group is None:
        group = click

    @group.command(**(command_kwargs or {}))
    @host_option
    @port_option
    @with_gunicorn_option
    @workers_option
    @verbose_option
    @debug_option
    def web(host: str, port: str, with_gunicorn: bool, workers: int, debug: bool):
        """Run the web application."""
        nonlocal app
        if isinstance(app, str):
            package_name, class_name = app.split(':')
            package = importlib.import_module(package_name)
            app = getattr(package, class_name)

        run_app(
            app=app,
            host=host,
            port=port,
            workers=workers,
            with_gunicorn=with_gunicorn,
            debug=debug,
        )

    return web


def run_app(
    app: 'flask.Flask',
    with_gunicorn: bool,
    host: Optional[str] = None,
    port: Optional[str] = None,
    workers: Optional[int] = None,
    debug: bool = False,
) -> NoReturn:
    """Run the application."""
    if not with_gunicorn:
        app.run(host=host, port=port, debug=debug)
    elif host is None or port is None or workers is None:
        raise ValueError('must specify host, port, and workers.')
    elif debug:
        raise ValueError('can not use debug=True with with_gunicorn=True')
    else:
        gunicorn_app = make_gunicorn_app(app, host, port, workers)
        gunicorn_app.run()


def make_gunicorn_app(
    app: 'flask.Flask',
    host: str,
    port: str,
    workers: int,
    **kwargs,
) -> 'gunicorn.app.base.BaseApplication':
    """Make a GUnicorn App."""
    from gunicorn.app.base import BaseApplication

    class StandaloneApplication(BaseApplication):
        def __init__(self, options=None):
            self.options = options or {}
            self.application = app
            super().__init__()

        def init(self, parser, opts, args):
            pass

        def load_config(self):
            for key, value in self.options.items():
                if key in self.cfg.settings and value is not None:
                    self.cfg.set(key.lower(), value)

        def load(self):
            return self.application

    kwargs.update({
        'bind': f'{host}:{port}',
        'workers': workers,
    })

    return StandaloneApplication(kwargs)



# -*- coding: utf-8 -*-

"""More click options."""

import logging
import multiprocessing
from typing import Union
from operator import itemgetter

import click

__all__ = [
    'verbose_option',
    'host_option',
    'port_option',
    'with_gunicorn_option',
    'workers_option',
    'force_option',
    'debug_option',
    'log_level_option',
]

LOG_FMT = '%(asctime)s %(levelname)-8s %(message)s'
LOG_DATEFMT = '%Y-%m-%d %H:%M:%S'


def _debug_callback(_ctx, _param, value):
    if not value:
        logging.basicConfig(level=logging.WARNING, format=LOG_FMT, datefmt=LOG_DATEFMT)
    elif value == 1:
        logging.basicConfig(level=logging.INFO, format=LOG_FMT, datefmt=LOG_DATEFMT)
    else:
        logging.basicConfig(level=logging.DEBUG, format=LOG_FMT, datefmt=LOG_DATEFMT)


verbose_option = click.option(
    '-v', '--verbose',
    count=True,
    callback=_debug_callback,
    expose_value=False,
    help="Enable verbose mode. More -v's means more verbose.",
)


def _number_of_workers() -> int:
    """Calculate the default number of workers."""
    return (multiprocessing.cpu_count() * 2) + 1


host_option = click.option('--host', type=str, default='0.0.0.0', help='Flask host.', show_default=True)
port_option = click.option('--port', type=int, default=5000, help='Flask port.', show_default=True)
with_gunicorn_option = click.option('--with-gunicorn', is_flag=True, help='Use gunicorn instead of flask dev server')
workers_option = click.option(
    '--workers',
    type=int,
    default=_number_of_workers(),
    help='Number of workers (when using --with-gunicorn)',
)
force_option = click.option('-f', '--force', is_flag=True)
debug_option = click.option('--debug', is_flag=True)

# sorted level names, by log-level
_level_names = sorted(logging._nameToLevel, key=logging._nameToLevel.get)

def log_level_option(default: Union[str, int] = logging.INFO):
    """Create a click option to select a log-level by name."""
    # normalize default to be a string
    if isinstance(default, int):
        default = logging.getLevelName(level=default)

    return click.option("-ll", "--log-level", type=click.Choice(choices=_level_names, case_sensitive=False), default=default)
